%%%-------------------------------------------------------------------
%%% File    : dialyzer.erl
%%% Author  : Tobias Lindahl <tobiasl@csd.uu.se>
%%%           Kostis Sagonas <kostis@it.uu.se>
%%% Description : The graphical user interface for the dialyzer.
%%%
%%% Created : 27 Apr 2004 by Tobias Lindahl <tobiasl@csd.uu.se>
%%%-------------------------------------------------------------------
-module(dialyzer).

-export([start/0]).

-record(state, {add_all, add_app, apps_box, analysis_pid, del_app, 
		clear_apps, clear_log, clear_warn,
		user_plt, dir_entry, file, gs, log, menu, mode,
		orig_CWD, packer, run, stop, top, fixpoint, warnings}).

-record(menu, {quit, empty_plt, overview, manual, about,
	       rebuild_plt, reinit_plt, save_log, save_warn}).

-record(mode, {start_from, application, fixpoint, module,
	       plt_app, plt_mod, plt_keep, qad}).


-record(analysis, {analysis_pid, current_app, files, file_type, fixpoint,
		   init_plt, mcg=none, plt_mode, printed_delim=false, 
		   user_plt}).

start() ->
  process_flag(trap_exit, true),

  GS = gs:start(),
  {ok, OriginalCWD} = file:get_cwd(),
  code:add_path(OriginalCWD),
  WH = [{width, 1000}, {height, 500}],

  %% --------- Top Window --------------
  TopWin = gs:window(GS, [{title, "Dialyzer v1.0"}, {configure, true},
			  {default, listbox, {bg, white}},
			  {default, editor, {bg, white}},
			  {default, entry, {bg, white}},
			  {default, button, {font, {helvetica, bold, 12}}},
			  {default, label, {font, {helvetica, bold, 12}}}
			  |WH]),
  Packer = gs:frame(TopWin, [{packer_x, [{stretch, 3},{fixed, 200},
					 {stretch, 7}]},
			     {packer_y, [{fixed, 25}, {fixed, 20},
					 {stretch, 1, 50},
					 {fixed, 25}, {fixed, 20},
					 {stretch, 1, 50},
					 {fixed, 25}]}]),

  %% --------- Application box --------------
  gs:label(Packer, [{label, {text, "Applications (or modules) to analyze"}}, 
		    {height, 20}, {pack_xy, {1,2}}]),
  AppsBox = gs:listbox(Packer, [{pack_xy, {1,3}},{vscroll, right}]),
  


  %% --------- File box --------------
  gs:label(Packer, [{label, {text, "File"}}, {height, 20}, {pack_xy, {1,5}}]),
  FilePacker = gs:frame(Packer, [{packer_x, [{fixed, 30}, {stretch, 1, 100}]},
				 {packer_y, [{fixed, 25}, {stretch, 1, 25}]},
				 {pack_xy, {1, 6}}]),
  gs:label(FilePacker, [{label, {text, "Dir:"}}, {pack_xy, {1,1}}]),
  DirEntry = gs:entry(FilePacker, [{height, 30}, {pack_xy, {2,1}},
				   {keypress, true}]),
  File = gs:listbox(FilePacker, [{pack_x, {1,2}}, {pack_y, 2},
				 {doubleclick, true}, {vscroll, right}]),

  %% --------- Options --------------
  gs:label(Packer, [{label, {text, "Analysis Options"}}, 
		    {height, 20}, {pack_xy, {2, 2}}]),
  ModePacker = gs:frame(Packer, [{packer_x, [{fixed, 75},{fixed, 120}]},
				 {packer_y, [{fixed, 15},
					     {fixed, 10}, % empty space
					     {fixed, 20}, {fixed, 20},
					     {fixed, 10}, % empty space
					     {fixed, 20}, {fixed, 20},
					     {fixed, 10}, % empty space
					     {fixed, 15}, {fixed, 25}, 
					     {fixed, 15}, {stretch, 1}]},
				 {bw, 10}, {relief, flat},
				 {default, {radiobutton, {align, w}}},
				 {default, {label, {align, w}}},
				 {pack_xy, {2, 3}}]),

  %% Bytecode vs. Source code
  gs:label(ModePacker, [{label, {text, "Analyze:"}},
			{height, 20}, {pack_xy, {1,1}}]),
  ByteCode = gs:radiobutton(ModePacker,
			    [{group, start_from},
			     {label, {text,"ByteCode"}},
			     {select, true},
			     {pack_xy, {2,1}}]),

  %% Analyze modules or applications
  gs:label(ModePacker, [{label, {text, "Granularity:"}},
			{height, 20}, {pack_xy, {1,3}}]),
  ModeFileModule = gs:radiobutton(ModePacker,
				  [{group, file_type},
				   {label, {text,"Module"}}, 
				   {pack_xy, {2,3}}]),
  ModeFileApp = gs:radiobutton(ModePacker,
			       [{group, file_type},
				{label, {text,"Application"}}, 
				{select, true},
				{pack_xy, {2,4}}]),

  %% Iteration Mode
  gs:label(ModePacker, [{label, {text, "Iteration:"}}, 
			{height, 20}, {pack_xy, {1,6}}]),
  ModeAnalysisQaD = gs:radiobutton(ModePacker,
				   [{group, analysis},
				    {label,{text,"One pass"}}, 
				    {pack_xy, {2,6}}]),
  ModeAnalysisFixpoint = gs:radiobutton(ModePacker,
					[{group, analysis},
					 {label,{text,"Fixpoint"}}, 
					 {select, true},
					 {pack_xy, {2,7}}]),

  %% Persistent Lookup Table Mode
  gs:label(ModePacker, [{label, {text, "Lookup Tab\nRe-init:"}},
			{height, 40}, {pack_xy, {1,10}},
			{anchor, nw}]),
			
  ModePLTModule = gs:radiobutton(ModePacker,
				 [{group, plt},
				  {label,{text,"Per Module"}}, 
				  {pack_xy, {2,9}}]),
  ModePLTApp = gs:radiobutton(ModePacker,
			      [{group, plt},
			       {label,{text,"Per Application"}}, 
			       {select, true},
			       {pack_xy, {2,10}}]),
  ModePLTKeep = gs:radiobutton(ModePacker,
			       [{group, plt},
				{label,{text,"Keep adding"}},
				{pack_xy, {2,11}}]),
  
  
  Mode = #mode{start_from=ByteCode,
	       application=ModeFileApp, 
	       fixpoint=ModeAnalysisFixpoint, 
	       module=ModeFileModule, 
	       plt_app=ModePLTApp, 
	       plt_mod=ModePLTModule, 
	       plt_keep=ModePLTKeep,
	       qad=ModeAnalysisQaD},

  handle_mode_change(Mode),
  
  %% --------- Log box --------------
  gs:label(Packer, [{label, {text, "Log"}}, {height, 20}, {pack_xy, {3,2}}]),
  Log = gs:editor(Packer, [{pack_x, 3}, {pack_y, 3}, {enable, false},
			   {font, {courier, 12}}, {vscroll, right},
			   {wrap, word}]),

  %% --------- Warnings box --------------
  gs:label(Packer, [{label, {text, "Warnings"}},{height, 20},{pack_xy, {3,5}}]),
  Warnings = gs:editor(Packer, [{pack_x, {2,3}}, {pack_y, 6} ,{enable, false},
				{font, {courier, 12}},{vscroll, right},
				{wrap, word}]),

  %% --------- Buttons --------------
  ButtonPackerHighLeft = 
    gs:frame(Packer, [{packer_x, [{fixed, 70},
				  {fixed, 70},
				  {stretch,1}]}, % empty space
		      {pack_xy, {1,4}}]),
  ButtonPackerHighRight = 
    gs:frame(Packer, [{packer_x, [{fixed, 70},
				  {fixed, 70},
				  {stretch,1}]}, % empty space
		      {pack_xy, {3,4}}]),
  ButtonPackerLowLeft = 
    gs:frame(Packer, [{packer_x, [{fixed, 70},
				  {fixed, 70},
				  {stretch,1}]}, % empty space
		      {pack_xy, {1,7}}]),
  ButtonPackerLowRight = 
    gs:frame(Packer, [{packer_x, [{fixed, 70},
				  {fixed, 70},
				  {stretch,1},% empty space
				  {fixed, 70},
				  {fixed, 70}]}, 
		      {pack_x, {2,3}}, {pack_y, 7}]),
  
  WHButton = [{width, 60},{height, 20}],
  AddApp = gs:button(ButtonPackerLowLeft, [{pack_xy, {1,1}}, 
					   {label,{text, "Add"}}|WHButton]),
  AddAllApps = gs:button(ButtonPackerLowLeft, [{pack_xy, {2,1}}, 
					       {label,{text, "Add All"}}
					       |WHButton]),
  DelApp = gs:button(ButtonPackerHighLeft, [{pack_xy, {1,1}}, 
					    {label,{text, "Delete"}}|WHButton]),
  ClearApps = gs:button(ButtonPackerHighLeft, [{pack_xy, {2,1}},
					       {label,{text, "Delete All"}}
					       |WHButton]),
  ClearLog = gs:button(ButtonPackerHighRight, [{pack_xy, {1,1}}, 
					       {label,{text, "Clear Log"}}
					       |WHButton]),
  ClearWarn = gs:button(ButtonPackerLowRight, [{pack_xy, {1,1}}, 
					       {label,{text, "Clear Warn"}}
					       |WHButton]),

  Run = gs:button(ButtonPackerLowRight, [{pack_xy, {4,1}},
					 {label,{text, "Run"}}|WHButton]),
  Stop = gs:button(ButtonPackerLowRight, [{pack_xy, {5,1}}, {enable, false}, 
					  {label,{text, "Stop"}}|WHButton]),



  %% --------- Menu --------------  
  MenuBar = gs:menubar(TopWin, []),

  %% File Menu
  MenuBarFile = gs:menubutton(MenuBar, [{label,{text,"File"}}]),
  MenuFile = gs:menu(MenuBarFile, []),
  MenuFileSaveWarn = gs:menuitem(MenuFile, [{label,{text,"Save Warnings"}}]),
  MenuFileSaveLog = gs:menuitem(MenuFile, [{label,{text,"Save Log"}}]),
  MenuFileQuit = gs:menuitem(MenuFile, [{label,{text,"Quit"}}]),

  %% PLT Menu
  MenuBarPLT = gs:menubutton(MenuBar, [{label,
					{text,"Plt"}}]),
  MenuPLT = gs:menu(MenuBarPLT, []),
  MenuPLTEmpty = gs:menuitem(MenuPLT, [{label,{text,"Init with empty plt"}},
				       {itemtype, check}, {select, false}]),
  MenuPLTRebuild = gs:menuitem(MenuPLT, [{label,
					  {text,"Re-build initial plt"}}]),
  MenuPLTReinit = gs:menuitem(MenuPLT, [{label,{text,"Re-init plt"}}]),
  

  %% Help
  MenuBarHelp = gs:menubutton(MenuBar, [{label,{text,"Help"}}, {side, right}]),
  MenuHelp = gs:menu(MenuBarHelp, []),
  MenuHelpOverview = gs:menuitem(MenuHelp, [{label,{text,"Overview"}}]),
  MenuHelpManual = gs:menuitem(MenuHelp, [{label,{text,"Manual"}}]),
  MenuHelpAbout = gs:menuitem(MenuHelp, [{label,{text,"About"}}]),


  Menu = #menu{quit=MenuFileQuit,
	       empty_plt=MenuPLTEmpty,
	       overview=MenuHelpOverview,
	       manual=MenuHelpManual,
	       about=MenuHelpAbout,
	       rebuild_plt=MenuPLTRebuild,
	       reinit_plt=MenuPLTReinit,
	       save_log=MenuFileSaveLog,
	       save_warn=MenuFileSaveWarn},
  
  %% --------- Init --------------
  gs:config(TopWin, {map, true}),
  gs:config(Packer, WH),

  User =
    case os:getenv("USER") of
      false ->      
	os:cmd("whoami");
      U -> U
    end,
  UserPlt = "/tmp/dialyzer_plt." ++ User,

  State = #state{add_all=AddAllApps,
		 add_app=AddApp, 
		 apps_box=AppsBox, 
		 del_app=DelApp, 
		 clear_apps=ClearApps, 
		 clear_log=ClearLog, 
		 clear_warn=ClearWarn, 
		 dir_entry=DirEntry,
		 file=File, 
		 gs=GS, 
		 log=Log, 
		 menu=Menu,
		 mode=Mode,
		 orig_CWD=OriginalCWD, 
		 packer=Packer, 
		 run=Run,
		 stop=Stop,
		 top=TopWin, 
		 user_plt=UserPlt,
		 warnings=Warnings},
  change_dir(".", File, DirEntry),
  
  Analysis = reinit_plt(State),
  loop(State, Analysis).


%% ----------------------------------------------------------------
%%
%%  Main Loop
%%


loop(State = #state{}, Analysis = #analysis{}) ->
  TopWin = State#state.top,
  Packer = State#state.packer,
  AppsBox = State#state.apps_box,
  File = State#state.file,
  DirEntry = State#state.dir_entry,
  Run = State#state.run,
  AddApp = State#state.add_app,
  AddAllApps = State#state.add_all,
  DelApp = State#state.del_app,
  ClearApps = State#state.clear_apps,
  ClearLog = State#state.clear_log,
  ClearWarn = State#state.clear_warn,
  Stop = State#state.stop,
  Log = State#state.log,

  AnalPid = Analysis#analysis.analysis_pid,

  Mode = State#state.mode,
  ModeFixpoint = Mode#mode.fixpoint,
  ModeQad = Mode#mode.qad,
  ModeFileApp = Mode#mode.application,
  ModeFileMod = Mode#mode.module,

  Menu = State#state.menu,
  Quit = Menu#menu.quit,
  Overview = Menu#menu.overview,
  Manual = Menu#menu.manual,
  About = Menu#menu.about,
  ReInitPlt = Menu#menu.reinit_plt,
  ReBuildPlt = Menu#menu.rebuild_plt,
  SaveLog = Menu#menu.save_log,
  SaveWarn = Menu#menu.save_warn,

  receive
    {gs, TopWin, configure, _Data, [W, H|_]} ->
      gs:config(Packer, [{width, W}, {height, H}]),
      loop(State, Analysis);
    {gs, TopWin, destroy, _Data, _Args} ->
      file:set_cwd(State#state.orig_CWD),
      ok;
    {gs, File, doubleclick, _, [_Id, Text|_]} ->
      change_dir(Text, File, DirEntry),
      loop(State, Analysis);
    {gs, DirEntry, keypress, _, ['Return'|_]} ->
      gs:config(TopWin, {setfocus, true}),
      change_dir(gs:read(DirEntry, text), File, DirEntry),
      loop(State, Analysis);
    {gs, DirEntry, keypress, _, _} ->
      loop(State, Analysis);
    %% ----- Buttons -----
    {gs, AddApp, click, _, _} ->
      case gs:read(Mode#mode.application, select) of
	true ->
	  handle_app_click(State);
	false ->
	  handle_mod_click(State)
      end,
      loop(State, Analysis);
    {gs, AddAllApps, click, _, _} ->
      handle_add_all_click(State, Mode),
      loop(State, Analysis);
    {gs, DelApp, click, _, _} ->
      handle_app_delete(State),
      loop(State, Analysis);
    {gs, ClearApps, click, _, _} ->
      gs:config(AppsBox, clear),
      loop(State, Analysis);
    {gs, ClearLog, click, _, _} ->
      Log = State#state.log,
      gs:config(Log, [{enable, true}]),
      gs:config(Log, clear),
      gs:config(Log, [{enable, false}]),
      loop(State, Analysis);
    {gs, ClearWarn, click, _, _} ->
      Warn = State#state.warnings,
      gs:config(Warn, [{enable, true}]),
      gs:config(Warn, clear),
      gs:config(Warn, [{enable, false}]),
      loop(State, Analysis);
    {gs, Run, click, _, _} ->
      NewAnalysis = start_analysis(State),
      loop(State, NewAnalysis);
    {gs, Stop, click, _, _} ->
      config_analysis_stop(State),
      exit(AnalPid, kill),
      loop(State, #analysis{});
    %% ----- Menu -----
    {gs, Quit, click, _, _} ->
      maybe_quit(State),
      loop(State, Analysis);
    {gs, Overview, click, _, _} ->
      overview(State),
      loop(State, Analysis);
    {gs, Manual, click, _, _} ->
      manual(State),
      loop(State, Analysis);
    {gs, About, click, _, _} ->
      about(State),
      loop(State, Analysis);
    {gs, ReBuildPlt, click, _, _} ->
      NewAnalysis = create_init_plt(State),
      loop(State, NewAnalysis);
    {gs, ReInitPlt, click, _, _} ->
      NewAnalysis = reinit_plt(State),
      loop(State, NewAnalysis);
    {gs, SaveLog, click, _, _} ->
      save_log(State),
      loop(State, Analysis);
    {gs, SaveWarn, click, _, _} ->
      save_warn(State),
      loop(State, Analysis);

    %% ----- Mode -----
    {gs, ModeFileMod, click, _, _} ->
      handle_mode_change(Mode),
      loop(State, Analysis);
    {gs, ModeFileApp, click, _, _} ->
      handle_mode_change(Mode),
      loop(State, Analysis);
    {gs, ModeFixpoint, click, _, _} ->
      handle_mode_change(Mode),
      loop(State, Analysis);
    {gs, ModeQad, click, _, _} ->
      handle_mode_change(Mode),
      loop(State, Analysis);

    %% ----- Analysis -----
    {AnalPid, current_app, NewApp} ->
      Delim = "\n==== Application: " ++ NewApp ++ "\n",
      update_editor(Log, Delim),
      loop(State, Analysis#analysis{current_app=NewApp});
    {AnalPid, log, LogMsg} ->
      update_editor(State#state.log, LogMsg),
      loop(State, Analysis);
    {AnalPid, warnings, Warnings} ->
      case Analysis#analysis.file_type of
	application ->
	  case Analysis#analysis.printed_delim of
	    false ->
	      Delim = "\n==== Application: " ++ Analysis#analysis.current_app 
		++ "\n",
	      update_editor(State#state.warnings, Delim ++ Warnings);
	    true ->
	      update_editor(State#state.warnings, Warnings)
	  end;
	module ->
	  update_editor(State#state.warnings, Warnings)
      end,
      loop(State, Analysis#analysis{printed_delim=true});
    {AnalPid, module_callgraph, MCG} ->
      io:format("Callgraph: ~p\n", [MCG]),
      loop(State, Analysis#analysis{mcg={module_callgraph, MCG}});
    {AnalPid, done} ->
      NewAnalysis = handle_analysis_done(State, Analysis),
      loop(State, NewAnalysis);
    Other ->
      io:format("Received ~p\n", [Other]),
      loop(State, Analysis)
  end.

%% ----------------------------------------------------------------
%%
%%  Main window actions
%%

change_dir(Text, File, DirEntry) ->
  file:set_cwd(Text),
  {ok, CWD} = file:get_cwd(),
  gs:config(DirEntry, [{text, CWD}]),
  {ok, List} = file:list_dir(CWD),
  gs:config(File, [{items, [".."|lists:sort(List)]}]),
  ok.

handle_app_click(S = #state{apps_box=AppsBox, file=File}) ->
  case gs:read(File, selection) of
    [] ->
      ok;
    List ->
      Add = [gs:read(File, {get, X}) || X <- List],
      {NoApps, RealApps} = filter_apps(Add),
      case NoApps of
	[] -> ok;
	_ -> 
	  error(S, "The following does not look like an application:\n" ++
		lists:flatten([[$\t|X]++"\n" || X<-NoApps]))
      end,
      add_files(RealApps, AppsBox, application)
  end.

filter_apps(Apps) ->
  filter_apps(Apps, [], []).

filter_apps([App|Left], NoApps, RealApps) ->
  case file:list_dir(App++"/ebin") of
    {ok, _} ->
      filter_apps(Left, NoApps, [App|RealApps]);
    {error, _} ->
      filter_apps(Left, [App|NoApps], RealApps)
  end;
filter_apps([], NoApps, RealApps) ->
  {NoApps, RealApps}.


handle_mod_click(S = #state{apps_box=AppsBox, file=File}) ->
  case gs:read(File, selection) of
    [] ->
      ok;
    List ->
      Add = [gs:read(File, {get, X}) || X <- List],
      {NoMod, RealMods} = filter_mods(Add),
      case NoMod of
	[] -> ok;
	_ -> 
	  error(S, "The following does not look like a module:\n" ++
		lists:flatten([[$\t|X]++"\n" || X<-NoMod]))
      end,
      add_files(RealMods, AppsBox, module)
  end.

filter_mods(Mods) ->
  Real = lists:filter(fun(X)->filename:extension(X)==".beam"end,Mods),
  Fake = lists:filter(fun(X)->not lists:member(X, Real)end, Mods),
  {Fake, Real}.

add_files(Add, AppsBox, Type) ->
  {ok, Cwd} = file:get_cwd(),
  RealAdd = ordsets:from_list([filename:join(Cwd, X) || X <- Add]),
  Set = gs:read(AppsBox, items),
  Filter =
    case Type of
      module -> fun(X)->filename:extension(X)==".beam"end;
      application -> fun(X)->filename:extension(X)/=".beam"end
    end,
  Set1 = ordsets:filter(Filter, Set),
  Apps = ordsets:union(RealAdd, Set1),
  gs:config(AppsBox, [{items, Apps}]),
  ok.

handle_add_all_click(#state{apps_box=AppsBox, file=File}, 
		     #mode{application=AppMode}) ->
  case gs:read(File, items) of
    [] ->
      ok;
    Set ->
      Set2 = ordsets:subtract(Set, [".."]),
      case gs:read(AppMode, select) of
	true ->
	  {_NoApps, RealApps} = filter_apps(Set2),
	  add_files(RealApps, AppsBox, application);
	false ->
	  {_NoMod, RealMods} = filter_mods(Set2),
	  add_files(RealMods, AppsBox, module)
      end
  end.


handle_app_delete(#state{apps_box=AppsBox})->
  case gs:read(AppsBox, selection) of
    [] ->
      ok;
    List ->
      [gs:config(AppsBox, {del, X}) || X <- List],
      ok
  end.

init_plt(State, From, To) ->
  case file:copy(From, To) of
    {ok, _} ->
      ok;
    {error, Reason} ->
      error(State,io_lib:format("Could not init lookup table:\n\n~w",[Reason])),
      error
  end.


config_analysis_start(State) ->
  gs:config(State#state.stop, {enable, true}),
  gs:config(State#state.run, {enable, false}),
  gs:config(State#state.del_app, {enable, false}),
  gs:config(State#state.clear_apps, {enable, false}),
  gs:config(State#state.add_app, {enable, false}),
  gs:config(State#state.add_all, {enable, false}),
  gs:config(State#state.clear_warn, {enable, false}),
  gs:config(State#state.clear_log, {enable, false}),

  Menu = State#state.menu,
  gs:config(Menu#menu.save_warn, {enable, false}),
  gs:config(Menu#menu.save_log, {enable, false}),
  gs:config(Menu#menu.reinit_plt, {enable, false}),
  gs:config(Menu#menu.rebuild_plt, {enable, false}),
  gs:config(Menu#menu.empty_plt, {enable, false}),

  Mode = State#state.mode,
  gs:config(Mode#mode.start_from, {enable, false}),
  gs:config(Mode#mode.application, {enable, false}),
  gs:config(Mode#mode.fixpoint, {enable, false}),
  gs:config(Mode#mode.module, {enable, false}),
  gs:config(Mode#mode.plt_app, {enable, false}),
  gs:config(Mode#mode.plt_mod, {enable, false}),
  gs:config(Mode#mode.plt_keep, {enable, false}),
  gs:config(Mode#mode.qad, {enable, false}).




config_analysis_stop(State) ->
  gs:config(State#state.stop, {enable, false}),
  gs:config(State#state.run, {enable, true}),
  gs:config(State#state.del_app, {enable, true}),
  gs:config(State#state.clear_apps, {enable, true}),
  gs:config(State#state.add_app, {enable, true}),
  gs:config(State#state.add_all, {enable, true}),
  gs:config(State#state.clear_warn, {enable, true}),
  gs:config(State#state.clear_log, {enable, true}),

  Menu = State#state.menu,
  gs:config(Menu#menu.save_warn, {enable, true}),
  gs:config(Menu#menu.save_log, {enable, true}),
  gs:config(Menu#menu.rebuild_plt, {enable, true}),
  gs:config(Menu#menu.reinit_plt, {enable, true}),
  gs:config(Menu#menu.empty_plt, {enable, true}),

  Mode = State#state.mode,
  gs:config(Mode#mode.start_from, {enable, true}),
  gs:config(Mode#mode.application, {enable, true}),
  gs:config(Mode#mode.fixpoint, {enable, true}),
  gs:config(Mode#mode.module, {enable, true}),
  gs:config(Mode#mode.plt_app, {enable, true}),
  gs:config(Mode#mode.plt_mod, {enable, true}),
  gs:config(Mode#mode.plt_keep, {enable, true}),
  gs:config(Mode#mode.qad, {enable, true}),
  handle_mode_change(Mode).


handle_mode_change(#mode{fixpoint=Fixpoint, module=FileMod,
			 plt_app=PltApp, plt_mod=PltMod}) ->
  case gs:read(FileMod, select) of
    true ->
      gs:config(PltMod, [{enable, true}]),
      case gs:read(PltApp, select) of
	true ->
	  gs:config(PltMod, [{select, true}]);
	false ->
	  ok
      end,
      gs:config(PltApp, [{enable, false}]);
    false ->
      gs:config(PltApp, [{enable, true}]),
      case gs:read(Fixpoint, select) of
	true ->
	  case  gs:read(PltMod, select) of
	    true ->
	      gs:config(PltApp, [{select, true}]);
	    false ->
	      ok
	  end,
	  gs:config(PltMod, [{enable, false}]);
	false ->
	  gs:config(PltMod, [{enable, true}])
      end
  end.
      
reinit_plt(State) ->
  OrigPlt = filename:join(os:getenv("DIALYZER_DIR"), "dialyzer_init_plt"),
  case file:read_file_info(OrigPlt) of
    {error, _} ->
      case dialog(State, 
		  "There is no persistent lookup table for OTP stdlib.\n"
		  "Do you want to create one now (takes a few minutes),\n"
		  "or do it later?",
		  "Now", "Later") of
	true -> 
	  create_init_plt(State);
	false ->
	  #analysis{}
      end;
    {ok, _} ->
      init_plt(State, OrigPlt, State#state.user_plt),
      #analysis{}
  end.



create_init_plt(State) ->  
  EmptyPlt = filename:join(os:getenv("DIALYZER_DIR"), "dialyzer_empty_plt"),
  Plt = filename:join(os:getenv("DIALYZER_DIR"), "dialyzer_init_plt"),
  case file:copy(EmptyPlt, Plt) of
    {error, Reason} ->
      error(State, io_lib:format("Could not init plt\n\n~p", [Reason])),
      #analysis{};
    {ok, _} ->
      Files = [filename:join(os:getenv("DIALYZER_OTP"), "lib/stdlib")],
      Analysis = #analysis{fixpoint=first, plt_mode=keep, file_type=application,
			   user_plt = Plt, files=Files},
      config_analysis_start(State),
      run_analysis(State, Analysis)
  end.


%% ----------------------------------------------------------------
%%
%%  Messages
%%

message(State, Message) ->
  message(State, "Dialyzer Message", Message).  

error(State, Message) ->
  message(State, "Dialyzer Error", Message).

message(#state{gs=GS, top=TopWin}, Title, Message) ->
  WH = [{width, 400}, {height, 100}],

  MessageWin = gs:window(GS, [{title, Title},
			      {default, button, {font, {helvetica, bold, 12}}}
			      |WH]),

  MessagePacker = gs:frame(MessageWin, [{packer_y, [{fixed, 75}, {fixed, 25}]},
					{packer_x, [{fixed, 175},{fixed, 50},
						    {fixed, 175}]}]),
  gs:label(MessagePacker, [{pack_x, {1,3}}, {pack_y, 1}, 
			   {label, {text, Message}}]),
  Ok = gs:button(MessagePacker, [{label, {text, "Ok"}}, 
				 {pack_xy, {2,2}}]),
  gs:config(MessageWin, {map, true}),
  gs:config(MessagePacker, WH),
  message_loop(Ok, MessageWin, TopWin).

message_loop(Ok, Win, TopWin) ->
  receive
    {gs, Ok, click, _, _} ->
      gs:destroy(Win);
    {gs, Win, destroy, _, _} ->
      ok;
    {gs, TopWin, destroy, _, _} ->
      exit(normal);
    _ ->
      message_loop(Ok, Win, TopWin)
  end.

dialog(S=#state{}, Message) ->
  dialog(S, Message, "Ok", "Cancel").

dialog(#state{gs=GS, top=TopWin}, Message, OkLabel, CancelLabel) ->
  WH = [{width, 400}, {height, 100}],
  WHButton = [{width, 70}, {height, 20}],

  DialogWin = gs:window(GS, [{title, "Dialyzer Message"},
			     {default, button, {font, {helvetica, bold, 12}}}
			     |WH]),
  DialogPacker = gs:frame(DialogWin, [{packer_y, [{fixed, 75}, {fixed, 25}]},
				      {packer_x, [{fixed, 150},{fixed, 50},
						  {fixed, 50},{fixed, 150}]}]),
  gs:label(DialogPacker, [{pack_x, {1,4}}, {pack_y, 1}, 
			  {label, {text, Message}}]),
  Ok = gs:button(DialogPacker, [{label, {text, OkLabel}}, 
				{pack_xy, {2,2}}|WHButton]),
  Cancel = gs:button(DialogPacker, [{label, {text, CancelLabel}}, 
				    {pack_xy, {3,2}}|WHButton]),

  gs:config(DialogWin, {map, true}),
  gs:config(DialogPacker, WH),
  dialog_loop(Ok, Cancel, DialogWin, TopWin).

dialog_loop(Ok, Cancel, Win, TopWin) ->
  receive
    {gs, Ok, click, _, _} ->
      gs:destroy(Win),
      true;
    {gs, Cancel, click, _, _} ->
      gs:destroy(Win),
      false;
    {gs, Win, destroy, _, _} ->
      false;
    {gs, TopWin, destroy, _, _} ->
      exit(normal);
    _ ->
      dialog_loop(Ok, Cancel, Win, TopWin)
  end.

maybe_quit(State=#state{top=TopWin,orig_CWD=OCWD})->
  case dialog(State, "Do you really want to quit?", "Yes", "No") of
    true ->
      file:set_cwd(OCWD),
      flush(),
      gs:destroy(TopWin),
      gs:stop(),
      exit(normal);
    false ->
      ok
  end.


%% ----------------------------------------------------------------
%%
%%  Menu actions
%%

overview(State) ->
  GS = State#state.gs,
  WH = [{width, 600}, {height, 360}],
  Win = gs:window(GS, [{title, "Dialyzer Overview"}, {configure, true},
		       {default, editor, {bg, white}} | WH]),
  Frame = gs:frame(Win, [{packer_x, [{stretch, 1}, {fixed, 60}, {stretch, 1}]}, 
			 {packer_y, [{stretch, 1}, {fixed, 30}]}
			 | WH]),
  Editor = gs:editor(Frame, [{pack_x, {1,3}}, {pack_y, 1},
			     {font, {courier, 12}}, {vscroll, right},
			     {wrap, word}]),
  Button = gs:button(Frame, [{label, {text, "Ok"}}, {pack_xy, {2,2}}]),
  gs:config(Win, {map, true}),
  gs:config(Frame, WH),

  AboutFile = filename:join(os:getenv("DIALYZER_DIR"), "doc/overview.txt"),
  case gs:config(Editor, {load, AboutFile}) of
    {error, Reason} ->
      gs:destroy(Win),
      error(State, 
	    io_lib:format("Could not find doc/overview.txt file!\n\n ~p", [Reason]));
    ok ->
      gs:config(Editor, {enable, false}),
      TopWin = State#state.top,
      receive
	{gs, Button, click, _, _} ->
	  gs:destroy(Win);
	{gs, TopWin, destroy, _, _} ->
	  exit(normal);
	{gs, Win, destroy, _, _} ->
	  ok
      end
  end.

manual(State) ->
  GS = State#state.gs,
  WH = [{width, 600}, {height, 500}],
  Win = gs:window(GS, [{title, "Dialyzer Manual"}, {configure, true},
		       {default, editor, {bg, white}} | WH]),
  Frame = gs:frame(Win, [{packer_x, [{stretch, 1}, {fixed, 60}, {stretch, 1}]}, 
			 {packer_y, [{stretch, 1}, {fixed, 30}]}
			 | WH]),
  Editor = gs:editor(Frame, [{pack_x, {1,3}}, {pack_y, 1},
			     {font, {courier, 12}}, {vscroll, right},
			     {wrap, word}]),
  Button = gs:button(Frame, [{label, {text, "Ok"}}, {pack_xy, {2,2}}]),
  gs:config(Win, {map, true}),
  gs:config(Frame, WH),

  AboutFile = filename:join(os:getenv("DIALYZER_DIR"), "doc/manual.txt"),
  case gs:config(Editor, {load, AboutFile}) of
    {error, Reason} ->
      gs:destroy(Win),
      error(State, 
	    io_lib:format("Could not find doc/manual.txt file!\n\n ~p", [Reason]));
    ok ->
      gs:config(Editor, {enable, false}),
      TopWin = State#state.top,
      receive
	{gs, Button, click, _, _} ->
	  gs:destroy(Win);
	{gs, TopWin, destroy, _, _} ->
	  exit(normal);
	{gs, Win, destroy, _, _} ->
	  ok
      end
  end.

about(State) ->
  GS = State#state.gs,
  WH = [{width, 600}, {height, 160}],
  Win = gs:window(GS, [{title, "About Dialyzer"}, {configure, true},
		       {default, editor, {bg, yellow}} | WH]),
  Frame = gs:frame(Win, [{packer_x, [{stretch, 1}, {fixed, 60}, {stretch, 1}]}, 
			 {packer_y, [{stretch, 1}, {fixed, 30}]}
			 | WH]),
  Editor = gs:editor(Frame, [{pack_x, {1,3}}, {pack_y, 1},
			     {font, {courier, 12}}, {vscroll, right},
			     {wrap, word}]),
  Button = gs:button(Frame, [{label, {text, "Ok"}}, {pack_xy, {2,2}}]),
  gs:config(Win, {map, true}),
  gs:config(Frame, WH),

  AboutFile = filename:join(os:getenv("DIALYZER_DIR"), "doc/about.txt"),
  case gs:config(Editor, {load, AboutFile}) of
    {error, Reason} ->
      gs:destroy(Win),
      error(State, 
	    io_lib:format("Could not find doc/about.txt file!\n\n ~p", [Reason]));
    ok ->
      gs:config(Editor, {enable, false}),
      TopWin = State#state.top,
      receive
	{gs, Button, click, _, _} ->
	  gs:destroy(Win);
	{gs, TopWin, destroy, _, _} ->
	  exit(normal);
	{gs, Win, destroy, _, _} ->
	  ok
      end
  end.


save_log(State) ->
  Log = State#state.log,
  {ok, CWD} = file:get_cwd(),
  {Win, Entry, OkButton, CancelButton} = file_box(State, "Save Log", CWD),
  save_loop(State, OkButton, CancelButton, Entry, Win, Log).

save_warn(State) ->
  Warn = State#state.warnings,
  {ok, CWD} = file:get_cwd(),
  {Win, Entry, OkButton, CancelButton} =
    file_box(State, "Save Warnings", CWD),
  save_loop(State, OkButton, CancelButton, Entry, Win, Warn).  

file_box(#state{gs=GS}, Title, Default) ->
  WH = [{width, 400}, {height, 75}],
  Win = gs:window(GS, [{title, Title}|WH]),
  WinPacker = gs:frame(Win, [{packer_y, [{fixed, 25}, {fixed, 25},
					 {fixed, 25}]},
			     {packer_x, [{fixed, 75},{fixed, 75},
					 {fixed, 75},
					 {fixed, 175}]}]),
  gs:label(WinPacker, [{pack_xy, {1,2}},
		       {label, {text, "Enter file:"}}]),
  Entry = gs:entry(WinPacker, [{pack_x, {2,4}}, 
			       {pack_y, 2}, {keypress, true}]),
  OkButton = gs:button(WinPacker, [{label, {text, "Ok"}}, 
				   {pack_xy, {2,3}}]),
  CancelButton = gs:button(WinPacker, [{label, {text, "Cancel"}}, 
				       {pack_xy, {3,3}}]),
  gs:config(Entry, {text, Default}),
  gs:config(Win, {map, true}),
  gs:config(WinPacker, WH),
  {Win, Entry, OkButton, CancelButton}.

save_loop(State, OkButton, CancelButton, Entry, Save, Editor) ->
  TopWin = State#state.top,
  receive
    {gs, OkButton, click, _, _} ->
      File = gs:read(Entry, text),
      case gs:config(Editor, {save, File}) of
	{error, _} ->
	  error(State, "Could not write to file "++File),
	  save_loop(State, OkButton, CancelButton, Entry, Save, Editor);
	_ ->
	  gs:destroy(Save)
      end;
    {gs, Entry, keypress, _, ['Return'|_]} ->
      File = gs:read(Entry, text),
      case gs:config(Editor, {save, File}) of
	{error, _} ->
	  error(State, "Could not write to file "++File),
	  save_loop(State, OkButton, CancelButton, Entry, Save, Editor);
	_ ->
	  gs:destroy(Save)
      end;
    {gs, Entry, keypress, _, _} ->
      save_loop(State, OkButton, CancelButton, Entry, Save, Editor);
    {gs, CancelButton, click, _, _} ->
      gs:destroy(Save);
    {gs, TopWin, destroy, _, _} ->
      exit(normal);
    {gs, Save, destroy, _, _} ->
      ok;
    _ ->
      save_loop(State, OkButton, CancelButton, Entry, Save, Editor)
  end.

%% ----------------------------------------------------------------
%%
%%  Run the analysis
%%

start_analysis(State = #state{mode=Mode, menu=Menu}) ->
  Analysis1 = 
    case gs:read(Mode#mode.fixpoint, select) of
      true -> #analysis{fixpoint=first};
      false -> #analysis{fixpoint=false}
    end,
  Analysis2 =
    case gs:read(Mode#mode.plt_keep, select) of
      true -> Analysis1#analysis{plt_mode=keep};
      false -> 
	case gs:read(Mode#mode.plt_app, select) of
	  true -> Analysis1#analysis{plt_mode=application};
	  false -> Analysis1#analysis{plt_mode=module}
	end
    end,
  Analysis3 =
    case gs:read(Mode#mode.application, select) of
      true -> Analysis2#analysis{file_type=application};
      false -> Analysis2#analysis{file_type=module}
    end,
  Analysis4 =
    case gs:read(Menu#menu.empty_plt, select) of
      true -> 
	From = filename:join(os:getenv("DIALYZER_DIR"), "dialyzer_empty_plt"),
	Analysis3#analysis{init_plt=From};
      false -> 
	From = filename:join(os:getenv("DIALYZER_DIR"), "dialyzer_init_plt"),
	Analysis3#analysis{init_plt=From}
    end,

  case check_analysis(Analysis4) of
    error ->
      Msg = "The currently selected mode is not valid!\n",
      error(State, Msg),
      #analysis{};
    ok ->
      case get_anal_files(State, Analysis4#analysis.file_type) of
	error ->
	  Msg = io_lib:format("You must choose one or more ~ws\n"
			      "before starting the analysis!\n",
			      [Analysis3#analysis.file_type]),
	  error(State, Msg),
	  config_analysis_stop(State),      
	  #analysis{};
	{ok, Files} ->
	  config_analysis_start(State),
	  run_analysis(State, Analysis4#analysis{files=Files, 
						 user_plt=State#state.user_plt})
      end
  end.

run_analysis(State, Analysis) ->
  MaybeInitPlt =
    case Analysis#analysis.plt_mode of
      keep ->
	ok;
      _ ->
	init_plt(State, Analysis#analysis.init_plt, 
		 Analysis#analysis.user_plt)
    end,
  case MaybeInitPlt of
    ok ->
      Self = self(),
      Fun = 
	fun()-> dialyzer_analysis:start(Analysis#analysis.files,
					Analysis#analysis.user_plt, 
					Self,
					Analysis#analysis.mcg,
					Analysis#analysis.fixpoint,
					Analysis#analysis.plt_mode,
					Analysis#analysis.file_type)end,
      Analysis#analysis{analysis_pid=spawn_link(Fun)};
    error ->
      #analysis{}
  end.

	    

handle_analysis_done(State, Analysis) ->
  case Analysis#analysis.fixpoint of
    first ->
      case Analysis#analysis.file_type of
	application ->
	  announce_second_run(State),
	  run_analysis(State, Analysis#analysis{fixpoint=second});
	module ->
	  message(State, "Analysis done"),
	  config_analysis_stop(State),
	  #analysis{}
      end;
    second ->
      message(State, "Analysis done"),
      config_analysis_stop(State),
      #analysis{};
    false ->
      case Analysis#analysis.plt_mode of
	module ->
	  #analysis{};
	_ ->
	  case dialog(State,"Analysis done\nProceed with fixpoint analysis?") of
	    true -> 
	      announce_second_run(State),
	      run_analysis(State, Analysis#analysis{fixpoint=second});
	    false -> 
	      config_analysis_stop(State),
	      #analysis{}
	  end
      end
  end.


get_anal_files(#state{apps_box=AppsBox}, FileType)->
  Filter =
    case FileType of
      module -> fun(X)->filename:extension(X)==".beam"end;
      application -> fun(X)->filename:extension(X)/=".beam"end
    end,
  case ordsets:filter(Filter, gs:read(AppsBox, items)) of
    [] -> error;
    Set -> {ok, Set}
  end.

flush() ->
  receive
    _ -> flush()
  after 
    0 -> ok
  end.

update_editor(Editor, Msg) ->
  gs:config(Editor, {enable, true}),
  NofRows = gs:read(Editor, size),
  gs:config(Editor, {insertpos, 'end'}),
  gs:config(Editor, {insert, {insert, Msg}}),
  NewNofRows = gs:read(Editor, size),
  ScrollPos = gs:read(Editor, vscrollpos),
  gs:config(Editor, {vscrollpos, ScrollPos + NewNofRows - NofRows}),
  gs:config(Editor, {enable, false}).

announce_second_run(#state{log=Log}) ->
  update_editor(Log, "\n========== Starting final run ==========\n").

check_analysis(#analysis{file_type=FileType,fixpoint=Fixpoint,user_plt=PLT}) ->
  case {FileType, Fixpoint, PLT} of
    {module, false, application} -> error;
    {module, first, application} -> error;
    {application, first, module} -> error;
    _ -> ok
  end.
